# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from cpython.ref cimport PyObject
from cpython cimport bool

from futures cimport FutureBase
cimport async_task
cimport scheduler
cimport _debug


cdef _debug.DebugOptions _debug_options

@cython.locals(task=async_task.AsyncTask, current_scheduler=scheduler.TaskScheduler, new_scheduler=scheduler.TaskScheduler)
cpdef object execute_on_separate_scheduler(object async_pull_fn, list contexts, tuple args=?, dict kwargs=?)

cdef inline object _await(tuple args)

@cython.locals(l=int)
cdef inline object _unwrap(tuple args)

cpdef result(object value)

# Private part

cdef _extract_tasks(object value, set result)




# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import async, result, cached
from .debug_cache import db, miss, mc, reset_cache_batch_indexes, flush_and_clear_local_caches


@cached(mc)
@async(pure=True)
def get_user(user_id):
    user = yield db.get('User: ' + str(user_id))
    result(None if user is miss else user); return


@cached(mc)
@async(pure=True)
def get_question(question_id):
    passport = yield db.get('Passport: ' + str(id))
    result(None if passport is miss else passport); return


@cached(mc)
@async(pure=True)
def get_question_user(question_id):
    question = yield get_question(question_id)
    user = None
    if question is not None:
        user = yield get_user(question['user'])
    result((question, user)); return


@async(pure=True)
def set_user(user_id, user):
    old_user = yield get_user(user_id)
    if old_user is not None:
        for qid in old_user['questions']:
            get_question_user.dirty(qid)
    if user is not None:
        for qid in user['questions']:
            get_question_user.dirty(qid)
    get_user.dirty(user_id)
    yield db.set('User: %i' % user_id, user)


@async(pure=True)
def set_question(question_id, question):
    get_question.dirty(question_id)
    get_question_user.dirty(question_id)
    yield db.set('Question: %i' % question_id, question)


# DB initialization

@async(pure=True)
def init():
    reset_cache_batch_indexes()
    print('init():')
    yield (
        set_question(1, {'title': 'Who?', 'user': 2}),
        set_question(2, {'title': 'Where?', 'user': 2}),
        set_question(3, {'title': 'When?', 'user': 2}),
        set_user(1, {'name': 'John Smith', 'questions': []}),
        set_user(2, {'name': 'James Bond', 'questions': [1, 2, 3]}),
    )

    flush_and_clear_local_caches()
    reset_cache_batch_indexes()
    print()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import async, await, result, ConstFuture, AsyncTaskResult, AsyncContext
from asynq.utils import execute_on_separate_scheduler
from qcore.asserts import assert_eq, assert_is, AssertRaises


class Context(AsyncContext):
    def __init__(self):
        self._is_active = False
        self.state = "pause"

    def resume(self):
        self.state = "resume"

    def pause(self):
        self.state = "pause"


def test_execute_on_separate_scheduler():
    ctx1 = Context()
    ctx2 = Context()
    ctx3 = Context()

    def inner(arg1, arg2=''):
        assert_eq('resume', ctx1.state)
        assert_eq('resume', ctx2.state)
        assert_eq('pause', ctx3.state)
        yield
        result(arg1 + arg2); return

    @async()
    def outer():
        with ctx3:
            pass

        with ctx2:
            value = execute_on_separate_scheduler(inner, [ctx1, ctx2], args=('hello',))
            assert_eq('hello', value)
            value = execute_on_separate_scheduler(inner, [ctx1, ctx2], args=('hello',),
                                                  kwargs={'arg2': ' world'})
            assert_eq('hello world', value)

    outer()


@async()
def f(x):
    return x * x


@async()
def g():
    res = yield
    assert_is(None, res)

    res = yield f.async(3)
    assert_eq(9, res)

    for sequence_length in range(6):
        res = yield tuple(f.async(i) for i in range(sequence_length))
        expected_tuple = tuple(i * i for i in range(sequence_length))
        assert_eq(expected_tuple, res)

        res = yield [f.async(i) for i in range(sequence_length)]
        assert_eq(list(expected_tuple), res)

        res = yield {i: f.async(i) for i in range(sequence_length)}
        expected_dict = {i: i * i for i in range(sequence_length)}
        assert_eq(expected_dict, res)


def test_unwrapping():
    g()


def test_result():
    with AssertRaises(AssertionError):
        result(ConstFuture(None))
    with AssertRaises(AsyncTaskResult):
        result(None)


def test_await():
    assert_is(None, await(None))
    assert_eq(9, await(f.async(3)))
    assert_eq((4, 9), await(f.async(2), f.async(3)))
    assert_eq((1, [4, 9]), await(f.async(1), [f.async(2), f.async(3)]))
    assert_eq(6, await(ConstFuture(6)))
    assert_eq({1: 2, 3: 4}, await({1: ConstFuture(2), 3: ConstFuture(4)}))

    with AssertRaises(TypeError):
        await(42)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import async
from .debug_cache import reset_caches
from .sample_model import get_question, get_question_user, get_user, init


@async(pure=True)
def async_test():
    print('Test:')
    question_ids = [1, 2, 3, 100]
    user_ids = [1, 2, 100]

    print('Questions and users:')
    for question, user in (yield tuple(map(get_question_user, question_ids))):
        print(('Question: %s, User: %s' % (str(question), str(user))))
    print()

#    print 'Questions:'
#    for question in (yield tuple(map(get_question, question_ids))):
#        print 'Question: %s' % (str(question))
#    print
#
#    print 'Users:'
#    for user in (yield tuple(map(get_user, user_ids))):
#        print 'User: %s' % (str(user))
#    print


def test():
    reset_caches()
    init()()
    async_test()()
    await()
#    flush_and_clear_local_caches()
#    async_test()()
#    flush_and_clear_local_caches()


if __name__ == '__main__':
    test()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore.asserts import AssertRaises, assert_eq
try:
    import cPickle as pickle
except ImportError:
    import pickle

from asynq import async
from .entities import unknown
from .helpers import Profiler
from .debug_cache import reset_caches
from . import model


def test():
    @async(pure=True)
    def async_test():
        model.reset_cache_batch_indexes()
        with Profiler('async_test()'):
            js, ay = yield model.User.get('JS'), model.User.get('AY')
            assert 1 == js.passport.id
            assert 2 == ay.passport.id

            jsp, ayp = yield js.passport, ay.passport

            assert jsp.name == 'John'
            assert ayp.name == 'Alex'

            # Changing user passports
            jsp.user = ay
            ayp.user = js
            js.passport = ayp
            ay.passport = jsp
            yield js, ay, jsp, ayp

            # Checking the result
            jsp, ayp = yield js.passport, ay.passport
            assert jsp.name == 'Alex'
            assert ayp.name == 'John'

            model.flush_and_clear_local_caches()
            assert_eq(8, model.mc._batch.index)
            assert_eq(11, model.db._batch.index)
        print()

    with Profiler('test_async_property()'):
        reset_caches()
        model.init()()
        async_test()()
    print()


def bad_cases_test():
    @async(pure=True)
    def async_test():
        with AssertRaises(AssertionError):
            model.User.new()
        with AssertRaises(AssertionError):
            model.User.new('AY', 'no matching field')
        with AssertRaises(AssertionError):
            model.User.new(None)
        with AssertRaises(AssertionError):
            model.User.new(unknown)

        ay = model.User.get('AY')
        with AssertRaises(AssertionError):
            ayp = yield ay.passport  # Key is unknown

        yield ay
        model.ayp = yield ay.passport

    reset_caches()
    model.init()()
    async_test()()


def pickle_test():
    # TODO: Fix this test
    return "Does not work w/Cython, TBD"

    @async(pure=True)
    def async_test():
        ay = yield model.User.get('AY')
        ayp = yield ay.passport
        ayp.name = 'Alexey'

        print('Before pickling:')
        print(ay, '\n', ayp)
        assert ayp.state.is_changed()

        ay, ayp = pickle_unpickle(ay, ayp)
        print('\nAfter pickling:')
        print(ay, '\n', ayp)
        assert ayp.state.is_changed()

        yield ay, ayp
        print('\nAfter sync:')
        print(ay, '\n', ayp)
        assert not ayp.state.is_changed()

    reset_caches()
    model.init()()
    async_test()()


def pickle_unpickle(*source):
    dump = pickle.dumps(source)
    return pickle.loads(dump)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import stderr, stdout
import time
import collections
import threading

import qcore.events as core_events


from . import debug
from . import futures
from . import batching
from . import _debug


assert str(__name__).endswith('asynq.scheduler') \
    or str(__name__).endswith('asynq.lib64.scheduler'), \
    "Are you importing asynq from the wrong directory?"


MAX_AWAIT_RECURSION_DEPTH = 10

_debug_options = _debug.options
_futures_none = futures._none


class AsyncTaskError(Exception):
    pass


class TaskScheduler(object):
    """Schedules and runs AsyncTask objects flushing batches when needed."""

    def __init__(self):
        self._last_dump_time = time.time()
        self._empty_set = set()  # Just a perf. optimization
        self.await_recursion_depth = 0
        self.max_await_recursion_depth = MAX_AWAIT_RECURSION_DEPTH
        self.on_before_batch_flush = core_events.EventHook()
        self.on_after_batch_flush = core_events.EventHook()
        self.reset()
        thread = threading.current_thread()
        thread_name = thread.name if thread.name else str(thread.ident)
        if '_state' in globals():
            _state.last_id += 1
            _id = _state.last_id
        else:
            _id = 0
        self.name = '%s / %d' % (thread_name, _id)

    def activate(self, must_await_on_exit, must_flush_batches_on_exit):
        return TaskSchedulerActivator(self, must_await_on_exit, must_flush_batches_on_exit)

    def reset(self):
        """Resets the scheduler to its initial state."""
        # Used to reverse the sequence of newly scheduled tasks
        # to execute them in correct order
        self._new_tasks = []
        self._tasks = collections.deque()
        self._batches = set()
        self.active_task = None

    def start(self, task, run_immediately=False):
        """Starts the specified task by simply scheduling it.
        Returns the task passed to this method.

        """
        if task.is_computed():
            return task
        if task.scheduler is not None:
            raise AsyncTaskError('Task is already scheduled.')
        if task.is_blocked():
            raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")

        if run_immediately:
            self._enqueue_new_tasks()
            self._schedule(task)
            self._continue_with_task()
        else:
            self._schedule(task)
        return task

    def continue_with(self, future, task):
        """Starts (or continues) the task after specified future gets computed.
        Returns the task passed to this method.

        """
        if task.is_computed():
            raise AsyncTaskError('Task is already computed.')
        if future.is_computed():
            if task.scheduler is None:
                return self.start(task)
            else:
                # Nothing to do in this case
                return task
        if task.scheduler is None:
            task.scheduler = self
        elif task.scheduler is not self:
            raise AsyncTaskError('Task is already scheduled on another scheduler.')
        self.add_dependency(task, future)
        self._try_schedule(task, True)
        return task

    def flush_batches(self, must_await=False):
        """Flushes all the batches that are currently 'known'
        to scheduler. Normally you shouldn't call this method.

        """
        if must_await:
            self.await()

        error = None
        while self._batches:
            batches = self._batches
            self._batches = set()
            for batch in batches:
                try:
                    if batch.items and not batch.is_flushed():
                        self._flush_batch(batch)
                except BaseException as e:
                    if error is None:
                        error = e
        if error is not None:
            raise error
        return 0

    def await(self, tasks=[]):
        """
        Executes a list of tasks ensuring all of them are completed when this method returns.

        :param tasks: tasks to await for
        :return: ``None``
        """
        self._await(tasks)

    def add_dependencies(self, task, yield_result):
        """Mark all futures in yield_result as dependencies for task.

        i.e. everything in yield_result needs to be computed before task can be continued.

        """
        if yield_result is None:  # Frequent case
            return
        for future in extract_futures(yield_result, []):
            self.add_dependency(task, future)

    def add_dependency(self, task, dependency):
        if dependency.is_computed():
            return
        if dependency in task._dependencies:
            if _debug_options.DUMP_DEPENDENCIES:
                debug.write('@async: +dependency: %s already depends on %s' %
                    (debug.str(task), debug.str(dependency)))
            return
        if _debug_options.DUMP_DEPENDENCIES:
            debug.write('@async: +dependency: %s needs %s' %
                (debug.str(task), debug.str(dependency)))
        dependency.make_dependency(task, self)

    def _await(self, tasks):
        active_task = self.active_task
        if active_task is not None:
            # Since task we await for can be (and most likely is)
            # a child of the current one, it already copied the
            # list of its contexts; so when we start it, these
            # contexts are going to be resumed, but we don't want
            # to resume already active contexts. That's why we
            # need to pause them.
            active_task._pause_contexts()
            # We must temporarily block active task in case
            # it's somehow (c) added to scheduler's queue
            blocking_dependency = _create_blocking_dependency()
            active_task._dependencies.add(blocking_dependency)
        try:
            # Must never throw an exception
            if len(tasks) == 0:
                # No arguments = run() must empty task queue
                return self._execute(False, self._empty_set)
            else:
                return self._execute(True, set(tasks))
        finally:
            if active_task is not None:
                active_task._dependencies.remove(blocking_dependency)
                active_task._resume_contexts()
        return 0

    def _execute(self, is_partial_execution, tasks):
        """Implements task execution loop."""
        # Must never throw an exception
        self.await_recursion_depth += 1
        current_active_task = self.active_task
        self.active_task = None
        try:
            if self.await_recursion_depth > self.max_await_recursion_depth:
                # We don't want to execute every next task on
                # stack in case all of these tasks invoke
                # await at some point
                dump_await_recursion = _debug_options.DUMP_AWAIT_RECURSION
                if dump_await_recursion:
                    debug.write("@async: await recursion: depth is %d, but the limit is %d" %
                                (self.await_recursion_depth, self.max_await_recursion_depth))
                if not is_partial_execution:
                    if dump_await_recursion:
                        debug.write("@async: await recursion: ignoring await()")
                    return 0
                self._try_execute_on_new_scheduler(tasks)
                # The rest part of tasks will be processed by this scheduler
            dump_scheduler_state = _debug_options.DUMP_SCHEDULER_STATE
            if is_partial_execution and len(tasks) == 0:
                return 0
            for task in tasks:
                creator = task.creator
                if creator is not None:
                    if task in creator._dependencies:
                        creator._remove_dependency(task)
                self._schedule(task)
            while True:  # Batch loop
                while True:  # Task loop
                    task = self._continue_with_task()
                    if dump_scheduler_state:
                        self.try_time_based_dump(task)
                    if task is None:
                        if _no_blocked_tasks(tasks):
                            return 0  # All specified tasks are computed
                        break  # No more tasks - let's try to flush the batch
                    if is_partial_execution and task.is_computed() and task in tasks:
                        tasks.remove(task)
                        if len(tasks) == 0:
                            return 0  # All specified tasks are computed
                try:
                    batch = self._continue_with_batch(is_partial_execution)
                    if batch is None:
                        return 0  # Nothing more to execute
                except AssertionError:
                    if _debug_options.DUMP_PRE_ERROR_STATE:
                        debug.write('Scheduler is awaiting for:')
                        for task in tasks:
                            task.dump(2)
                        debug.write('----------------------------------------------------------------------\n')
                        raise
        finally:
            self.active_task = current_active_task
            self.await_recursion_depth -= 1
        return 0

    def _try_execute_on_new_scheduler(self, tasks):
        new_scheduler_tasks = []
        for task in tasks:
            if task.scheduler is None:
                new_scheduler_tasks.append(task)
        for task in new_scheduler_tasks:
            tasks.remove(task)

        if _debug_options.DUMP_AWAIT_RECURSION:
            debug.write("@async: await recursion: going to run %d tasks on a new scheduler" % len(new_scheduler_tasks))

        if len(new_scheduler_tasks) == 0:
            pass
        else:
            new_scheduler = self._clone()
            try:
                with new_scheduler.activate(True, False):
                    new_scheduler.await(new_scheduler_tasks)
            finally:
                self._accept_changes(new_scheduler)

    def schedule(self, task):
        self._schedule(task)

    def _schedule(self, task):
        if task.is_blocked():
            if _debug_options.DUMP_SCHEDULE_TASK:
                debug.write("@async: can't schedule blocked task %s" % debug.str(task))
            assert task.scheduler is self, \
                'Task %s is already scheduled on another scheduler.' % debug.str(task)
            return False
        if task.is_computed():
            if _debug_options.DUMP_SCHEDULE_TASK:
                debug.write("@async: can't schedule computed task %s" % debug.str(task))
            return False
        return self._schedule_without_checks(task)

    def _schedule_without_checks(self, task):
        if _debug_options.DUMP_SCHEDULE_TASK:
            debug.write('@async: scheduling task %s' % debug.str(task))

        if task.scheduler is None:
            task.scheduler = self
        else:
            assert task.scheduler is self, 'Task is already scheduled on another scheduler.'

        if not task.is_scheduled:
            task.is_scheduled = True
            self._new_tasks.append(task)
        return True

    def _try_schedule(self, task, enqueue_new_tasks):
        if task.is_computed() or task.is_blocked():
            return False
        if enqueue_new_tasks:
            self._enqueue_new_tasks()
        return self._schedule_without_checks(task)

    def schedule_batch(self, batch):
        self._schedule_batch(batch)

    def _schedule_batch(self, batch):
        if batch.is_flushed():
            if _debug_options.DUMP_SCHEDULE_BATCH:
                debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
            return False
        if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:
            debug.write('@async: scheduling batch %s' % debug.str(batch))
        self._batches.add(batch)
        return True

    def _flush_batch(self, batch):
        self.on_before_batch_flush(batch)
        try:
            batch.flush()
        finally:
            self.on_after_batch_flush(batch)
        return 0

    def _enqueue_new_tasks(self):
        l = len(self._new_tasks)
        if l == 0:
            pass
        else:
            i = 0
            while i < l:
                self._tasks.append(self._new_tasks[i])
                i += 1
            del self._new_tasks[:]

    def _continue_with_task(self):
        """Continues the next task.
        Returns the task that made some progress or None.

        """
        # Queuing new tasks
        self._enqueue_new_tasks()
        dump_continue_task = _debug_options.DUMP_CONTINUE_TASK
        while len(self._tasks) != 0:
            task = self._tasks.popleft()
            task.is_scheduled = False
            if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster
                if dump_continue_task:
                    debug.write('@async: skipping %s' % debug.str(task))
                continue

            # Executing the task
            old_active_task = self.active_task
            self.active_task = task
            try:
                if dump_continue_task:
                    debug.write('@async: -> continuing %s' % debug.str(task))
                task._continue()
            finally:
                # task._continue should never throw an error, but
                # it's better to implement active task switch in safe way
                self.active_task = old_active_task
                # No need to reschedule: task._continue() ensures this
                # anyway
                if dump_continue_task:
                    debug.write('@async: <- continued %s' % debug.str(task))
            return task
        if dump_continue_task:
            debug.write('@async: no task to continue')
        return None

    def _continue_with_batch(self, assert_no_batch=True):
        """
        Flushes one of batches (the longest one by default).

        :param assert_no_batch: indicates whether exception must be
                                raised if there is no batch to flush
        :return: the batch that was flushed, if there was a flush;
                 otherwise, ``None``.

        """
        batch = self._select_batch_to_flush()
        if batch is None:
            if _debug_options.DUMP_FLUSH_BATCH:
                debug.write('@async: no batch to flush')
            if assert_no_batch:
                debug.dump(self)
                assert False, \
                    "Can't proceed further: no task to continue or batch to flush, active tasks: " \
                    "%s" % self._tasks
            else:
                return None
        self._batches.remove(batch)
        self._flush_batch(batch)
        return batch

    def _select_batch_to_flush(self):
        """Returns the batch having highest priority,
        or ``None``, if there are no batches.

        This method uses ``BatchBase.get_priority()`` to
        determine the priority.

        Side effect: this method removed flushed batches.

        :return: selected batch or None.

        """
        best_batch = None
        best_priority = None
        batches_to_remove = None
        for batch in self._batches:
            if not batch.items or batch.is_flushed():
                if batches_to_remove is None:
                    batches_to_remove = [batch]
                else:
                    batches_to_remove.append(batch)
                continue
            priority = batch.get_priority()
            if best_batch is None or best_priority < priority:
                best_batch = batch
                best_priority = priority
        if batches_to_remove:
            for batch in batches_to_remove:
                self._batches.remove(batch)
        return best_batch

    def _clone(self):
        """Protected method cloning the scheduler.

        This method is used to create a new scheduler temporarily
        replacing the current one to handle await recursion

        """
        clone = type(self)()
        clone._last_dump_time = self._last_dump_time
        clone.max_await_recursion_depth = self.max_await_recursion_depth
        clone.on_before_batch_flush = self.on_before_batch_flush
        clone.on_after_batch_flush = self.on_after_batch_flush
        return clone

    def _accept_changes(self, clone):
        """Protected method accepting the changes made to this scheduler's clone."""

        self._last_dump_time = clone._last_dump_time
        self.max_await_recursion_depth = clone.max_await_recursion_depth
        self.on_before_batch_flush = clone.on_before_batch_flush
        self.on_after_batch_flush = clone.on_after_batch_flush
        for batch in clone._batches:
            self._batches.add(batch)

    def __str__(self):
        return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \
            (type(self), repr(self.name),
             len(self._tasks), len(self._new_tasks), len(self._batches),
             str(self.active_task))

    def __repr__(self):
        return self.__str__()

    def dump(self, indent=0):
        debug.write(debug.str(self), indent)
        if self._tasks:
            debug.write('Task queue:', indent + 1)
            for task in self._tasks:
                task.dump(indent + 2)
        else:
            debug.write('No tasks in task queue.', indent + 1)
        if self._new_tasks:
            debug.write('New tasks:', indent + 1)
            for task in self._new_tasks:
                task.dump(indent + 2)
        if self._batches:
            debug.write('Batches:', indent + 1)
            for batch in self._batches:
                batch.dump(indent + 2)

    def try_time_based_dump(self, last_task=None):
        current_time = time.time()
        if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:
            return
        self._last_dump_time = current_time
        debug.write('\n--- Scheduler state dump: --------------------------------------------')
        try:
            self.dump()
            if last_task is not None:
                debug.write('Last task: %s' % debug.str(last_task), 1)
        finally:
            debug.write('----------------------------------------------------------------------\n')
            stdout.flush()
            stderr.flush()


class TaskSchedulerActivator(object):
    def __init__(self, scheduler, must_await_on_exit, must_flush_batches_on_exit):
        self._scheduler = scheduler
        self._old_scheduler = None
        self._must_await_on_exit = must_await_on_exit
        self._must_flush_batches_on_exit = must_flush_batches_on_exit

    def __enter__(self):
        if _debug_options.DUMP_SCHEDULER_CHANGE:
            debug.write("@async: -> scheduler activation")
        self._old_scheduler = get_scheduler()
        set_scheduler(self._scheduler)

    def __exit__(self, type, value, traceback):
        try:
            try:
                if self._must_await_on_exit:
                    self._scheduler.await()
            finally:
                if self._must_flush_batches_on_exit:
                    self._scheduler.flush_batches()
        finally:
            set_scheduler(self._old_scheduler)
            if _debug_options.DUMP_SCHEDULER_CHANGE:
                debug.write("@async: <- scheduler activation")


class LocalTaskSchedulerState(threading.local):
    def __init__(self):
        self.last_id = 0
        self.reset()

    def reset(self):
        self.current = TaskScheduler()

_state = LocalTaskSchedulerState()
globals()['_state'] = _state


def get_scheduler():
    global _state
    return _state.current


def reset():
    _state.reset()


def set_scheduler(scheduler):
    global _state
    _state.current = scheduler
    if _debug_options.DUMP_SCHEDULER_CHANGE:
        debug.write("@async: scheduler = %s" % debug.str(scheduler))


def get_active_task():
    global _state
    s = _state.current
    return None if s is None else s.active_task


def extract_futures(value, result):
    """Enumerates all the futures inside a particular value."""
    if value is None:
        pass
    elif isinstance(value, futures.FutureBase):
        result.append(value)
    elif type(value) is tuple or type(value) is list:
        for item in value:
            extract_futures(item, result)
    return result


def _no_blocked_tasks(tasks):
    computed = [task for task in tasks if task.is_computed()]
    for task in computed:
        if task in tasks:
            tasks.remove(task)
    return len(tasks) == 0


def _create_blocking_dependency():
    return futures.ConstFuture("async_blocking_dependency")



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import async, await
from .debug_cache import reset_caches
from .model import reset_cache_batch_indexes, get_user_passport, Profiler, flush_and_clear_local_caches, mc, db, init


def test():
    @async(pure=True)
    def test1():
        reset_cache_batch_indexes()
        with Profiler('test1()'):
            (js, jsp), (ay, ayp), none = yield (
                get_user_passport('JS'),
                get_user_passport('AY'),
                get_user_passport('none'))

            assert js['passport_id'] == 1
            assert ay['passport_id'] == 2
            assert jsp['name'] == 'John'
            assert ayp['name'] == 'Alex'
            assert none is None

            flush_and_clear_local_caches()
            assert mc._batch.index == 5
            assert db._batch.index == 3
        print()

    @async(pure=True)
    def test2():
        reset_cache_batch_indexes()
        with Profiler('test2()'):
            user_ids = ['AY', 'JS']
            for user, passport in (yield tuple(map(get_user_passport, user_ids))):
                assert user['passport_id'] != None
                assert passport['user_id'] != None

            flush_and_clear_local_caches()
            assert mc._batch.index == 2
            assert db._batch.index == 1
        print()

    with Profiler('test_batching()'):
        reset_caches()
        init()()
        test1()()
        test2().start()()  # Actually the same as simply test2()()
    print()




# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import qcore
from qcore.asserts import assert_eq
from asynq import async, result, cached
from .debug_cache import db, mc, miss, reset_cache_batch_indexes, flush_and_clear_local_caches
from .helpers import Profiler
from .entities import Entity, EntityField, PersisterBase


@cached(mc)
@async(pure=True)
def get_user(id):
    user = yield db.get('User: ' + id)
    if user is miss:
        result(None); return
    else:
        user['id'] = id
        result(user); return


@cached(mc)
@async(pure=True)
def get_passport(id):
    passport = yield db.get('Passport: ' + str(id))
    if passport is miss:
        result(None); return
    else:
        passport['id'] = id
        result(passport); return


@cached(mc)
@async(pure=True)
def get_user_passport(user_id):
    user = yield get_user(user_id)
    if user is None:
        result(None); return
    else:
        passport = yield get_passport(user['passport_id'])
        result((user, passport)); return


@async(pure=True)
def set_user(id, user):
    if 'id' in user:
        del user['id']
    get_user.dirty(id)
    get_user_passport.dirty(id)
    yield db.set('User: ' + id, user)


@async(pure=True)
def set_passport(id, passport):
    if 'id' in passport:
        del passport['id']
    get_passport.dirty(id)
    if passport is not None:
        get_user_passport.dirty(id, passport['user_id'])
    yield db.set('Passport: ' + str(id), passport)


class SimplePersister(PersisterBase):
    def __init__(self, get_method, set_method):
        self.get_method = get_method
        self.set_method = set_method

    @async(pure=True)
    def create(self, entity):
        yield self.update(entity, True)

    @async(pure=True)
    def load(self, entity):
        key = self.get_key(entity)
        data = yield self.get_method(key)
        if data is None:
            raise KeyError(
                "%s with key=%r does not exist." %
                (qcore.inspection.get_full_name(type(entity)), key))
        entity.update_data(data, mark_changed=False)

    @async(pure=True)
    def update(self, entity, create=False):
        print('update %r' % entity)
        key = self.get_key(entity)
        data = yield self.get_method(key)
        if create:
            if data is not None:
                raise KeyError(
                    "%s with key=%r already exists." %
                    (qcore.inspection.get_full_name(type(entity)), key))
            data = {}
        else:
            if data is None:
                raise KeyError(
                    "%s with key=%r does not exist." %
                    (qcore.inspection.get_full_name(type(entity)), key))
        data.update(entity.get_data(changes_only=True))
        yield self.set_method(key, data)

    @async(pure=True)
    def delete(self, entity):
        key = self.get_key(entity)
        data = yield self.get_method(key)
        if data is None:
            raise KeyError(
                "%s with id=%r does not exist." %
                (qcore.inspection.get_full_name(type(entity)), key))
        yield self.set_method(key, data)

    def get_key(self, entity):
        key = entity.key
        if len(key) == 1:
            key = key[0]
        return str(key)


class User(Entity):
    persister = SimplePersister(get_user, set_user)
    id = EntityField().key()
    passport_id = EntityField()

    @property
    def passport(self):
        return Passport.get(self.passport_id)

    @passport.setter
    def passport(self, value):
        self.passport_id = None if value is None else value.id


class Passport(Entity):
    persister = SimplePersister(get_passport, set_passport)
    id = EntityField().key()
    name = EntityField()
    surname = EntityField()
    user_id = EntityField()

    @property
    def user(self):
        return User.get(self.user_id)

    @user.setter
    def user(self, value):
        self.user_id = None if value is None else value.id


# No-yield methods

@cached(mc)
def get_user_passport_no_yield(user_id):
    user = get_user(user_id)()
    if user is None:
        return None
    else:
        passport = get_passport(user['passport_id'])()
        return (user, passport)


# DB initialization

@async(pure=True)
def init():
    reset_cache_batch_indexes()
    with Profiler('init()'):
        js = User.new('JS')
        js_p = Passport.new(1)
        js_p.name = 'John'
        js_p.surname = 'Smith'
        js.passport = js_p
        js_p.user = js

        ay = User.new('AY')
        ay_p = Passport.new(2)
        ay_p.name = 'Alex'
        ay_p.surname = 'Yakunin'
        ay.passport = ay_p
        ay_p.user = ay

        print(js)
        print(js_p)

        yield js, ay, js_p, ay_p

        print(js)
        print(js_p)

        flush_and_clear_local_caches()
        assert_eq(9, mc._batch.index)
        assert_eq(9, db._batch.index)
    print()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from cpython.ref cimport PyObject

cimport qcore.events as core_events
cimport futures
cimport async_task
cimport batching
cimport _debug


cdef _debug.DebugOptions _debug_options
cdef object _futures_none


cdef class TaskScheduler(object):
    cdef public set _empty_set
    cdef public float _last_dump_time
    cdef public list _new_tasks
    cdef public object _tasks
    cdef public set _batches
    cdef public int await_recursion_depth
    cdef public int max_await_recursion_depth
    cdef public core_events.EventHook on_before_batch_flush
    cdef public core_events.EventHook on_after_batch_flush
    cdef public async_task.AsyncTask active_task
    cdef public str name

    cpdef TaskSchedulerActivator activate(self, bint must_await_on_exit, bint must_flush_batches_on_exit)
    cpdef reset(self)

    cpdef async_task.AsyncTask start(self, async_task.AsyncTask task,
            bint run_immediately=?)
    cpdef async_task.AsyncTask continue_with(self, futures.FutureBase future, async_task.AsyncTask task)
    cpdef int flush_batches(self, bint must_await=?) except -1

    cpdef add_dependencies(self, async_task.AsyncTask task, object yield_result)

    cpdef add_dependency(self, async_task.AsyncTask task, futures.FutureBase dependency)
    cpdef schedule(self, async_task.AsyncTask task)
    cpdef schedule_batch(self, batching.BatchBase batch)

    @cython.locals(l=int, active_task=async_task.AsyncTask)
    cpdef int _await(self, list tasks) except -1
    @cython.locals(dump_scheduler_state=bint, dump_await_recursion=bint,
                   task=async_task.AsyncTask, creator=async_task.AsyncTask,
                   captured_dependencies=list)
    cpdef int _execute(self, bint is_partial_execution, set tasks) except -1
    @cython.locals(new_scheduler=TaskScheduler, new_scheduler_tasks=list,
                   task=async_task.AsyncTask)
    cpdef _try_execute_on_new_scheduler(self, set tasks)

    cdef bint _schedule(self, async_task.AsyncTask task) except -1
    cdef bint _schedule_without_checks(self, async_task.AsyncTask task) except -1
    cdef bint _try_schedule(self, async_task.AsyncTask task, bint enqueue_new_tasks) except -1
    cdef bint _schedule_batch(self, batching.BatchBase batch) except -1
    cdef int _flush_batch(self, batching.BatchBase batch) except -1

    @cython.locals(l=int, i=int)
    cdef inline object _enqueue_new_tasks(self)
    @cython.locals(dump_continue_task=bint, task=async_task.AsyncTask)
    cdef async_task.AsyncTask _continue_with_task(self)
    cdef batching.BatchBase _continue_with_batch(self, bint assert_no_batch=?)

    cpdef batching.BatchBase _select_batch_to_flush(self)

    @cython.locals(clone=TaskScheduler)
    cpdef object _clone(self)
    cpdef _accept_changes(self, TaskScheduler clone)

    cpdef dump(self, int indent=?)
    @cython.locals(current_time=float)
    cpdef try_time_based_dump(self, object last_task=?)

cdef class TaskSchedulerActivator(object):
    cdef TaskScheduler _scheduler
    cdef TaskScheduler _old_scheduler
    cdef bint _must_await_on_exit
    cdef bint _must_flush_batches_on_exit

cpdef object extract_futures(object value, list result)

cdef object _state

cpdef TaskScheduler get_scheduler()
cpdef set_scheduler(scheduler)

cpdef async_task.AsyncTask get_active_task()
cpdef bint _no_blocked_tasks(tasks) except -1

cpdef futures.FutureBase _create_blocking_dependency()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import qcore
import inspect
import six
from six.moves import xrange
from asynq import async, async_proxy, ConstFuture, result, FutureBase


unknown = qcore.MarkerObject(u'unknown @ entities')
must_load = qcore.MarkerObject(u'must_load @ entities')


class EntityState(qcore.Flags):
    # Base flags
    new = 1
    deleted = 2
    loaded = 0x10
    changed = 0x20
    # Few useful combinations
    must_load = 0
    new_must_save = new | loaded | changed
    deleted_saved = loaded | deleted
    deleted_must_save = loaded | deleted | changed

    # A few useful check methods

    def is_new(self):
        return EntityState.new in self

    def is_deleted(self):
        return EntityState.deleted in self

    def is_loaded(self):
        return EntityState.loaded in self

    def is_changed(self):
        return EntityState.changed in self

    def is_synchronized(self):
        return \
            EntityState.loaded in self and \
            EntityState.changed not in self


class PersisterBase(object):
    @async_proxy(pure=True)
    def sync(self, entity):
        # print 'Syncing: %r' % entity
        state = entity.state
        if not state.is_loaded():
            assert not state.is_changed(), \
                'Cannot sync: the entity is not yet loaded, but changed.'
            task = self.load(entity)
            task.on_computed.subscribe(lambda _: entity._mark_synchronized())
            return task
        else:
            return self.save(entity)

    @async_proxy(pure=True)
    def save(self, entity):
        state = entity.state
        if not state.is_changed():
            return ConstFuture(None)
        if state.is_new():
            task = self.create(entity)
        elif state.is_deleted():
            task = self.delete(entity)
        else:
            task = self.update(entity)
        task.on_computed.subscribe(lambda _: entity._mark_synchronized())
        return task

    @async(pure=True)
    def create(self, entity):
        # Must raise an exception if entity cannot be created
        # (e.g. if it already exists in the storage)
        raise NotImplementedError()

    @async(pure=True)
    def load(self, entity):
        # Must raise an exception if entity cannot be loaded
        # (e.g. if it absents in the storage)
        raise NotImplementedError()

    @async(pure=True)
    def update(self, entity):
        # Must raise an exception if entity cannot be saved
        # (e.g. if it absents in the storage)
        raise NotImplementedError()

    @async(pure=True)
    def delete(self, entity):
        # Must raise an exception if entity cannot be deleted
        # (e.g. if it absents in the storage)
        raise NotImplementedError()


class NotImplementedPersister(PersisterBase):
    pass


class NoPersister(PersisterBase):
    @async(pure=True)
    def create(self, entity):
        entity._mark_synchronized()

    @async(pure=True)
    def load(self, entity):
        entity._mark_synchronized()

    @async(pure=True)
    def update(self, entity):
        entity._mark_synchronized()

    @async(pure=True)
    def delete(self, entity):
        entity._mark_synchronized()


class EntityField(object):
    cls = None
    name = None
    internal_name = None
    backing_name = None
    is_key = False
    key_index = None
    is_lazy = False

    def __init__(self, default=None, name=None, backing_name=None):
        self.default = default
        self.set_name(name, backing_name=backing_name)

    def set_name(self, name, backing_name=None):
        self.name = name
        self.internal_name = None if name is None else '_p_' + name
        self.backing_name = backing_name or name
        return self

    def key(self, index=0):
        self.is_key = index is not None
        self.key_index = index
        return self

    def lazy(self, lazy=True):
        self.is_lazy = lazy
        return self

    def is_own(self, cls):
        return self.cls is cls

    def is_inherited(self, cls):
        return self.cls is not cls

    def __get__(self, instance, owner):
        if instance is None:
            return self  # Handles cls.property case
        result = instance.__dict__.get(self.name, unknown)
        if result is unknown and instance.state.is_new():
            return self.default
        return result

    def __set__(self, instance, value):
        assert not self.is_key, "Key field %r cannot be changed." % self.name
        if instance.track_changes:
            d = instance.__dict__
            old_value = instance.__dict__.get(self.name, unknown)
            if old_value is unknown and instance.state.is_new():
                old_value = self.default
            d[self.name] = value
            if old_value != value:
                instance._mark_changed()
                if 'changes' in d:
                    instance.changes.add(self.name)
                else:
                    instance.changes = set((self.name,))
        else:
            instance.__dict__[self.name] = value
            instance._mark_changed()

    def set_unsafe(self, instance, value):
        instance.__dict__[self.name] = value

    def __repr__(self):
        result = "%s(%r, name=%r" % (
            self.__class__.__name__, self.default, self.name)
        if self.backing_name != self.name:
            result += ', backing_name=%r' % self.backing_name
        result += ')'
        if self.is_key:
            result += '.key(%r)' % self.key_index
        if self.is_lazy:
            result += '.lazy()'
        return result


class EntityMetadata(object):
    def __init__(self, cls):
        self.cls = cls
        self.all_fields = []
        self.key_fields = []
        self.non_key_fields = []
        self.fields = {}
        self.field_indexes = {}

        for name, f in list(inspect.getmembers(cls)):
            if not isinstance(f, EntityField):
                continue
            if f.cls is None:
                f.cls = cls
            if f.name is None:
                f.set_name(name)
            else:
                assert name == f.name, \
                    'EntityField name mismatch: %r != %r (class: %s).' % \
                    (name, f.name, qcore.inspection.get_full_name(cls))

            assert name not in self.fields, \
                'Duplicate property name: %s (class: %s).' % \
                (name, qcore.inspection.get_full_name(cls))
            self.fields[f.name] = f

            if f.is_key:
                self.key_fields.append(f)
            else:
                self.non_key_fields.append(f)

        self.key_fields = sorted(self.key_fields, key=lambda f: f.key_index)

        for i in xrange(0, len(self.key_fields)):
            f = self.key_fields[i]
            assert i == f.key_index, \
                'Key index mismatch for property %s (class: %s): %r != %r.' % \
                (f.name, qcore.inspection.get_full_name(cls), i, f.key_index)
            assert not f.is_lazy, \
                'Key property cannot be lazy: %s (class: %s).' % \
                (f.name, qcore.inspection.get_full_name(cls))

        self.all_fields = self.key_fields + self.non_key_fields
        for i in xrange(0, len(self.all_fields)):
            f = self.all_fields[i]
            self.field_indexes[f.name] = i

    def __repr__(self):
        def list_fields(fields):
            result = ''
            for f in fields:
                result += '\n    %r' % f
                if f.is_inherited(self.cls):
                    result += ' # Inherited'
            return result

        result = '%s(%s):' % (
            self.__class__.__name__,
            qcore.inspection.get_full_name(self.cls))
        if self.key_fields:
            result += '\n  Key fields:'
            result += list_fields(self.key_fields)
            result += '\n  Other fields:'
            result += list_fields(self.non_key_fields)
        return result


class EntityType(type):
    def __init__(cls, what, bases=None, dict=None):
        super(EntityType, cls).__init__(what, bases, dict)
        cls.metadata = EntityMetadata(cls)


class Entity(six.with_metaclass(EntityType, FutureBase)):
    persister = NotImplementedPersister()
    metadata = None  # Set by metaclass

    changes = None
    track_changes = False

    def __init__(self, key, direct_call=True):
        super(Entity, self).__init__()
        key_fields = self.metadata.key_fields
        key_field_count = len(key_fields)
        assert direct_call is False, \
            "Entity.__init__ cannot be called directly. " \
            "Use Entity.get(...) or Entity.new(...) instead."
        assert key_field_count > 0, \
            "Key is not defined for %s." % \
            qcore.inspection.get_full_name(self.__class__)
        assert key_field_count == len(key), \
            "Key length (%i) differ from the expected one (%i)." % (
                len(key), key_field_count)
        self._state = EntityState.must_load
        for i in xrange(0, key_field_count):
            f = key_fields[i]
            v = key[i]
            assert not (v is None or v is unknown), \
                "Key field %r cannot be %r." % (f.name, v)
            f.set_unsafe(self, v)

    @classmethod
    def get(cls, *key):
        return cls(key, direct_call=False)

    @classmethod
    def new(cls, *key):
        return cls(key, direct_call=False)._mark_new()

    @property
    def state(self):
        return self._state

    @property
    def key(self):
        return tuple([
            getattr(self, f.name)
            for f in self.metadata.key_fields
        ])

    @property
    def full_key(self):
        return (type(self),) + self.key

    def update(self, **values):
        """A helper method allowing to set multiple fields at once."""
        for k, v in values.items():
            setattr(self, k, v)
        return self

    def delete(self):
        return self._mark_deleted()

    def get_data(self, changes_only=False):
        data = {}
        if changes_only:
            if not self.state.is_changed():
                return data
            if not self.track_changes:
                changes_only = False
        if changes_only:
            changes = self.changes or qcore.empty_tuple
            for f in self.metadata.non_key_fields:
                v = getattr(self, f.name)
                if v is not unknown and f.name in changes:
                    data[f.name] = v
        else:
            for f in self.metadata.non_key_fields:
                v = getattr(self, f.name)
                if v is not unknown:
                    data[f.name] = v
        return data

    def update_data(self, data, mark_changed=True):
        if mark_changed:
            self._mark_changed()
        fields = self.metadata.fields
        for k, v in data.items():
            f = fields.get(k)
            if f is None or f.is_key:
                continue
            f.set_unsafe(self, v)

    @async(pure=True)
    def sync(self):
        ret = yield self.persister.sync(self)
        result(ret); return

    def _compute(self):
        if not self.is_computed():
            self.set_value(self)  # Prevents the recursion
        return self.sync()()

    def _mark_new(self):
        self._state = EntityState.new_must_save
        if 'changes' in self.__dict__:
            del self.changes
        self.reset_unsafe()
        return self

    def _mark_changed(self):
        assert self._state.is_loaded(), "The entity must be loaded first."
        self._state |= EntityState.changed
        self.reset_unsafe()
        return self

    def _mark_deleted(self):
        assert self._state.is_loaded(), "The entity must be loaded first."
        self._state |= EntityState.deleted_must_save
        self.reset_unsafe()
        return self

    def _mark_synchronized(self):
        if self._state.is_deleted():
            self._state = self.state.deleted_saved
        else:
            self._state = self.state.loaded
        if 'changes' in self.__dict__:
            del self.changes
        return self

    def __getstate(self):
        return self.__dict__

    def __setstate__(self, state):
        # Needed to properly unpickle entities
        self.__dict__ = state
        if self._state.is_changed():
            self.reset_unsafe()

    def __repr__(self):
        result = '%s(%s):' % (
            self.__class__.__name__,
            ', '.join([repr(f) for f in self.key]))
        result += '\n  State: %s' % self.state
        if self.track_changes and self.changes:
            result += ' (%s)' % ', '.join(self.changes)
        for k, v in self.get_data().items():
            result += '\n  %s: %r' % (k, v)
        return result



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from .futures import FutureBase
from . import async_task
from . import scheduler
from . import _debug


_debug_options = _debug.options


def execute_on_separate_scheduler(async_pull_fn, contexts, args=(), kwargs={}):
    current_scheduler = scheduler.get_scheduler()
    new_scheduler = current_scheduler._clone()
    try:
        with new_scheduler.activate(True, False):
            task = async_task.AsyncTask(async_pull_fn(*args, **kwargs), async_pull_fn, args, kwargs)
            # The above tasks should have no contexts right now (because it is created in a vacuum).
            # We filter the contexts that are already active because they will remain active
            # throughout the execution of this task. Not filtering them would result in calling
            # resume on them twice in a row (which most async_contexts do not support).
            task._contexts = [ctx for ctx in contexts if not ctx._is_active]
            new_scheduler.await([task])
    finally:
        current_scheduler._accept_changes(new_scheduler)
    assert task.is_computed(), 'the task should have been computed at this point'
    return task.value()


def await(*args):
    """Runs all the tasks specified in args,
    and finally returns args unwrapped.

    """
    return _await(args)


def _await(args):
    tasks = set()
    _extract_tasks(args, tasks)
    scheduler.get_scheduler()._execute(True, tasks)
    return _unwrap(args)


def _unwrap(args):
    l = len(args)
    assert l != 0, "No arguments to unwrap."
    if l == 1:
        return async_task.unwrap(args[0])  # Special case: no need to return tuple here
    return async_task.unwrap(args)


def result(value):
    """An async equivalent for return for async methods: result(x); return."""
    assert not isinstance(value, FutureBase), \
        "You probably forgot to yield this value before returning"
    raise async_task.AsyncTaskResult(value)


# Private part

def _extract_tasks(value, result):
    if value is None:
        pass
    elif isinstance(value, async_task.AsyncTask):
        result.add(value)
    elif type(value) is tuple or type(value) is list:
        for item in value:
            _extract_tasks(item, result)
    elif type(value) is dict:
        for item in six.itervalues(value):
            _extract_tasks(item, result)
    elif isinstance(value, FutureBase):
        pass
    else:
        raise TypeError(
            "Cannot process an object of type '%s': only futures and None are allowed."
            % type(value)
        )
